/*
 * Copyright (C) 2001 Cedric Duval <cedricduval@free.fr>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include "mutt.h"
#include "mutt_menu.h"
#include "mapping.h"
#include "mutt_curses.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>

#define SIG_DISPLAY_LINES 4
#define SEPARATOR(x) ((*x == '\n') || (mutt_strcmp (x, "-- \n") == 0))
#define SIG_ADD_LINE(x,y) mutt_add_list (x, strtok (y, "\n"))

extern char* SearchBuffers[MENU_MAX];

typedef struct sig_list
{
  struct sig_list *next;
  LIST *sig;
} SIG_LIST;

typedef struct sig_dir
{
  struct sig_dir *next;
  char *name;
} SIG_DIR;

typedef LIST * ENTRY;

typedef struct entry_dir
{
  int tagged;
  SIG_DIR *data;
} ENTRY_DIR;

static struct mapping_t SigHelp[] = {
  { N_("Exit"),   OP_EXIT },
  { N_("Search"), OP_SEARCH },
  { N_("Random"), OP_RANDOM_SIG },
  { N_("Help"),   OP_HELP },
  { NULL }
};

static struct mapping_t SigDirHelp[] = {
  { N_("Exit"),   OP_EXIT },
  { N_("Search signature"), OP_SIG_SEARCH },
  { N_("Help"),   OP_HELP },
  { NULL }
};

void menu_next_entry (MUTTMENU *menu);
void menu_prev_entry (MUTTMENU *menu);


static int sig_match (LIST *s, regex_t *re)
{
  while (s)
  {
    if (regexec (re, s->data, (size_t)0, NULL, (int)0) == 0)
      return 1;
    s = s->next;
  }
  return 0;
}

static void read_sig_file (char *name, SIG_LIST **begin, regex_t *re)
{
  FILE *fp;
  char buf[STRING];
  LIST *sig = NULL;
  SIG_LIST *first, *cur;
  int append = 0;

  if (!(fp = safe_fopen (name, "r")))
  {
    mutt_error (_("Can't open signature file %s"), name);
    return;
  }

  for (first = *begin; first && first->next; first = first->next, append++)
    ; /* append results to an existing list */
  cur = first;

  while (fgets (buf, sizeof (buf), fp))
  {
    if (buf[0] && !SEPARATOR (buf))
    {
      sig = SIG_ADD_LINE (NULL, buf);

      while (fgets (buf, sizeof (buf), fp) && buf[0] && !SEPARATOR (buf))
	SIG_ADD_LINE (sig, buf);

      if (re && !sig_match (sig, re))
	mutt_free_list (&sig); /* previous sig didn't match the regexp */
      else
      {
	/* add signature */
	if (first == NULL)
	  first = cur = (SIG_LIST *) safe_calloc (1, sizeof (SIG_LIST));
	else
	{
	  cur->next = (SIG_LIST *) safe_calloc (1, sizeof (SIG_LIST));
	  cur = cur->next;
	}

	cur->sig = sig;
	cur->next = NULL;
      }
    }
  }

  if (!append)
    *begin = first;

  safe_fclose (&fp);
}

static void sig_make_entry (char *s, size_t slen, MUTTMENU *menu, int num)
{
  ENTRY *table = (ENTRY *) menu->data;

  snprintf (s, slen, "%3d %s",
	    num + 1,
	    table[num]->data);
}

static int sig_menu_search (MUTTMENU *menu, regex_t *re, int num)
{
  return (sig_match (((ENTRY *)menu->data)[num], re) ? 0 : REG_NOMATCH);
}

static void draw_sig_frame (LIST *s)
{
  int i;

  for (i = 1; i <= SIG_DISPLAY_LINES; i++)
  {
    if (s)
    {
      mvaddstr (i, 0, s->data);
      s = s->next;
    }
    else
      move (i, 0);

    clrtoeol ();
  }

  SETCOLOR (MT_COLOR_STATUS);
  mvaddstr (SIG_DISPLAY_LINES + 1, 0, _("-- Signature"));
  BKGDSET (MT_COLOR_STATUS);
  clrtoeol ();

  BKGDSET (MT_COLOR_NORMAL);
  SETCOLOR (MT_COLOR_NORMAL);
}

static void free_sig_list (SIG_LIST **sigs)
{
  SIG_LIST *cur;

  while (*sigs)
  {
    cur = *sigs;
    *sigs = (*sigs)->next;
    mutt_free_list (&cur->sig);
    safe_free ((void **) (void *) &cur);
  }
}

static void append_signature (char *msg_file, LIST *s)
{
  FILE *fp;

  if ((fp = safe_fopen (msg_file, "a")) == 0)
    mutt_perror (msg_file);
  else
  {
    if (option (OPTSIGDASHES))
      fputs ("\n-- \n", fp);

    for (; s; s = s->next)
      fprintf (fp, "%s\n", s->data);

    mutt_message (_("Signature appended to your mail"));
    safe_fclose (&fp);
  }
}

static LIST *sig_list_menu (char *file, SIG_LIST *list)
{
  LIST *result = NULL;
  SIG_LIST *sigl;
  MUTTMENU *menu;
  ENTRY *SigTable;
  char helpstr[SHORT_STRING], title[SHORT_STRING];
  int i, done = 0;

  snprintf (title, sizeof (title), _("Signature : %s"), file);

  menu = mutt_new_menu (MENU_SIG);
  menu->make_entry = sig_make_entry;
  menu->tag = NULL;
  menu->search = sig_menu_search;
  menu->title = title;
  menu->help = mutt_compile_help (helpstr, sizeof (helpstr),
				  MENU_SIG, SigHelp);
  menu->offset = SIG_DISPLAY_LINES + 2;
  menu->pagelen = LINES - SIG_DISPLAY_LINES - 4;
  
  for (sigl = list; sigl; sigl = sigl->next)
    menu->max++;

  menu->data = SigTable = (ENTRY *) safe_calloc (menu->max, sizeof (ENTRY));

  for (i = 0, sigl = list; sigl; i++, sigl = sigl->next)
    SigTable[i] = sigl->sig;

  while (!done)
  {
    switch (mutt_menuLoop (menu))
    {
      case OP_GENERIC_SELECT_ENTRY:
        result = SigTable[menu->current];
	done = 1;
	break;

      case OP_PREV_SIG:
	menu_prev_entry (menu);
	draw_sig_frame (SigTable[menu->current]);
	break;

      case OP_NEXT_SIG:
	menu_next_entry (menu);
	draw_sig_frame (SigTable[menu->current]);
	break;

      case OP_REDRAW:
	menu->offset = SIG_DISPLAY_LINES + 2;
	menu->pagelen = LINES - SIG_DISPLAY_LINES - 4;
	draw_sig_frame (SigTable[menu->current]);
	break;

      case OP_RANDOM_SIG:
	menu->current = LRAND () % menu->max;
	draw_sig_frame (SigTable[menu->current]);
	menu->redraw |= REDRAW_MOTION;
	break;

      case OP_EXIT:
	set_option (OPTNEEDREDRAW);
        done = 1;
        break;
    }
  }
  
  mutt_menuDestroy (&menu);
  safe_free ((void **) (void *) &SigTable);
  return result;
}

static SIG_LIST *sig_search_filter (MUTTMENU *menu, char *path)
{
  regex_t re;
  char buf[STRING];
  SIG_LIST *result = NULL;
  int i;
  char* searchBuf = menu->menu >= 0 && menu->menu < MENU_MAX ?
                    SearchBuffers[menu->menu] : NULL;

  strfcpy (buf, searchBuf ? searchBuf : "", sizeof(buf));
  if (mutt_get_field (_("Search for: "), buf,
		      sizeof (buf), M_CLEAR) != 0 || !buf[0])
    return (NULL);
  if (menu->menu >= 0 && menu->menu < MENU_MAX)
  {
    mutt_str_replace (&SearchBuffers[menu->menu], buf);
    searchBuf = SearchBuffers[menu->menu];
  }

  if ((i = regcomp (&re, searchBuf, REG_NOSUB | REG_EXTENDED | REG_WORDS
		    | mutt_which_case (searchBuf))) != 0)
  {
    regerror (i, &re, buf, sizeof (buf));
    regfree (&re);
    mutt_error ("%s", buf);
    return (NULL);
  }

  /* building list of sigs matching the regexp */
  for (i = 0; i < menu->max; i++)
  {
    /* search in every file if none is tagged */
    if (((ENTRY_DIR *) menu->data)[i].tagged || (menu->tagged == 0))
    {
      snprintf (buf, sizeof (buf), "%s/%s", path,
		((ENTRY_DIR *) menu->data)[i].data->name);
      read_sig_file (buf, &result, &re);
    }
  }

  regfree (&re);
  if (!result)
    mutt_error (_("Not found."));

  return (result);
}

/* returns the list of files in this directory */
static SIG_DIR *sig_directory (char *path)
{
  DIR *dp;
  struct dirent *de;
  struct stat s;
  SIG_DIR *first = NULL, *cur = NULL;
  char file[_POSIX_PATH_MAX + SHORT_STRING];

  if ((dp = opendir (path)) == NULL)
  {
    mutt_perror (path);
    return (NULL);
  }
  
  while ((de = readdir (dp)))
  {
    if ((de->d_name)[0] == '.') /* no hidden files */
      continue;

    snprintf (file, sizeof (file), "%s/%s", path, de->d_name);
    if (lstat (file, &s) == -1)
      continue;

    if ((!S_ISREG (s.st_mode)) && (!S_ISLNK (s.st_mode)))
      continue;

    if (first == NULL)
      cur = first = safe_calloc (1, sizeof (SIG_DIR));
    else
    {
      cur->next = safe_calloc (1, sizeof (SIG_DIR));
      cur = cur->next;
    }
    cur->name = safe_strdup (de->d_name);
    cur->next = NULL;
  }
  closedir (dp);
  return first;
}

static void sig_dir_make_entry (char *s, size_t slen, MUTTMENU *menu, int num)
{
  ENTRY_DIR *table = (ENTRY_DIR *) menu->data;

  snprintf (s, slen, "%c %3d - %s",
	    table[num].tagged ? '*' : ' ',
	    num + 1,
	    table[num].data->name);
}

static int sig_dir_tag (MUTTMENU *menu, int n, int m)
{
  ENTRY_DIR *cur = &((ENTRY_DIR *) menu->data)[n];
  int ot = cur->tagged;
  
  cur->tagged = m >= 0 ? m : !cur->tagged;
  return cur->tagged - ot;

}

static int sig_dir_sort (const void *a, const void *b)
{
  ENTRY_DIR *pa = (ENTRY_DIR *) a;
  ENTRY_DIR *pb = (ENTRY_DIR *) b;

  return (mutt_strcmp (pa->data->name, pb->data->name));
}

static int sig_dir_menu (char *path, char *msg_file)
{
  MUTTMENU *menu;
  SIG_LIST *sigl;
  LIST *result = NULL;
  ENTRY_DIR *FileTable;
  SIG_DIR *list, *files;
  char buf[STRING], helpstr[SHORT_STRING], title[SHORT_STRING];
  int i, done = 0;

  if ((list = sig_directory (path)) == NULL)
    return -1;

  snprintf (title, sizeof (title), "Signature directory : %s", path);

  menu = mutt_new_menu (MENU_SIG_DIR);
  menu->make_entry = sig_dir_make_entry;
  menu->search = NULL; /* search within files with sig_search_filter() */
  menu->tag = sig_dir_tag;
  menu->title = title;
  menu->help = mutt_compile_help (helpstr, sizeof (helpstr),
				  MENU_SIG_DIR, SigDirHelp);

  for (files = list; files; files = files->next)
    menu->max++;

  menu->data = FileTable = (ENTRY_DIR *) safe_calloc (menu->max,
						      sizeof (ENTRY_DIR));

  for (i = 0, files = list; files; i++, files = files->next)
    FileTable[i].data = files;

  qsort (FileTable, menu->max, sizeof (ENTRY_DIR), sig_dir_sort);

  while (!done)
  {
    switch (mutt_menuLoop (menu))
    {
      case OP_SIG_SEARCH:
	sigl = sig_search_filter (menu, path);

	if (sigl)
	{
	  if ((result = sig_list_menu (_("query results"), sigl)) != NULL)
	  {
	    append_signature (msg_file, result);
	    done = 1;
	  }

	  MAYBE_REDRAW (menu->redraw);
	  free_sig_list (&sigl);
	}
	break;

      case OP_GENERIC_SELECT_ENTRY:
	snprintf (buf, sizeof (buf), "%s/%s", path,
		  FileTable[menu->current].data->name);
	sigl = NULL;
	read_sig_file (buf, &sigl, NULL);

	if (sigl)
	{
	  if ((result = sig_list_menu (buf, sigl)) != NULL)
	  {
	    append_signature (msg_file, result);
	    done = 1;
	  }

	  MAYBE_REDRAW (menu->redraw);
	  free_sig_list (&sigl);
      	}
	break;

      case OP_EXIT:
        done = 1;
        break;
    }
  }

  while (list)
  {
    safe_free ((void **) (void *) &list->name);
    files = list;
    list = list->next;
    safe_free ((void **) (void *) &files);
  }
  safe_free ((void **) (void *) &FileTable);
  mutt_menuDestroy (&menu);
  return 0;
}

void mutt_signature (char *msg_file)
{
  if (!SigDirectory)
  {
    mutt_error (_("variable 'signatures_directory' is unset"));
    return;
  }

  if (sig_dir_menu (SigDirectory, msg_file) == -1)
    mutt_error (_("%s: no files in this directory"), SigDirectory);
  else
    set_option (OPTNEEDREDRAW);
}
